@@ -26,12 +26,13 @@
 #include "posix_to_bcachefs.h"
 #include "libbcachefs.h"
 #include "crypto.h"
-#include "libbcachefs/errcode.h"
-#include "libbcachefs/opts.h"
-#include "libbcachefs/super-io.h"
-#include "libbcachefs/util.h"
 
-#include "libbcachefs/darray.h"
+#include "errcode.h"
+#include "opts.h"
+#include "init/fs.h"
+#include "sb/io.h"
+#include "util/util.h"
+#include "util/darray.h"
 
 #define OPTS						\
 x(0,	replicas,		required_argument)	\
@@ -60,27 +61,27 @@ static void format_usage(void)
 
 	bch2_opts_usage(OPT_FORMAT|OPT_FS);
 
-	puts("      --replicas=#            Sets both data and metadata replicas\n"
-	     "      --encrypted             Enable whole filesystem encryption (chacha20/poly1305)\n"
-	     "      --passphrase_file=file  File containing passphrase used for encryption/decryption\n"
-	     "      --no_passphrase         Don't encrypt master encryption key\n"
+	puts("      --replicas=#             Sets both data and metadata replicas\n"
+	     "      --encrypted              Enable whole filesystem encryption (chacha20/poly1305)\n"
+	     "      --passphrase_file=file   File containing passphrase used for encryption/decryption\n"
+	     "      --no_passphrase          Don't encrypt master encryption key\n"
 	     "  -L, --fs_label=label\n"
 	     "  -U, --uuid=uuid\n"
 	     "      --superblock_size=size\n"
-	     "      --version=version       Create filesystem with specified on disk format version instead of the latest\n"
-	     "      --source=path           Initialize the bcachefs filesystem from this root directory\n"
+	     "      --version=version        Create filesystem with specified on disk format version instead of the latest\n"
+	     "      --source=path            Initialize the bcachefs filesystem from this root directory\n"
 	     "\n"
 	     "Device specific options:");
 
 	bch2_opts_usage(OPT_FORMAT|OPT_DEVICE);
 
-	puts("      --fs_size=size          Size of filesystem on device\n"
-	     "  -l, --label=label           Disk label\n"
+	puts("      --fs_size=size           Size of filesystem on device\n"
+	     "  -l, --label=label            Disk label\n"
 	     "\n"
 	     "  -f, --force\n"
-	     "  -q, --quiet                 Only print errors\n"
-	     "  -v, --verbose               Verbose filesystem initialization\n"
-	     "  -h, --help                  Display this help and exit\n"
+	     "  -q, --quiet                  Only print errors\n"
+	     "  -v, --verbose                Verbose filesystem initialization\n"
+	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Device specific options must come before corresponding devices, e.g.\n"
 	     "  bcachefs format --label cache /dev/sdb /dev/sdc\n"
@@ -322,7 +323,7 @@ int cmd_format(int argc, char *argv[])
 		if (opts.source)
 			build_fs(c, opts.source);
 
-		bch2_fs_stop(c);
+		bch2_fs_exit(c);
 	}
 	bch2_opt_strs_free(&fs_opt_strs);
 	darray_exit(&devices);
