@@ -111,6 +111,7 @@ static int move_btree_pred(struct btree_trans *trans, void *_arg,
 	if (!args->move_alloc && btree_id_is_alloc(btree))
 		return false;
 
+	data_opts->write_flags |= BCH_WRITE_only_specified_devs;
 	return true;
 	return k.k->type == KEY_TYPE_btree_ptr_v2 && !btree_id_is_alloc(btree);
 }
@@ -379,8 +380,9 @@ static int finish_image(struct bch_fs *c,
 	 * sb->nr_devices must be 1 so that it can be mounted without UUID
 	 * conflicts
 	 */
+	struct bch_sb_field_members_v2 *mi = bch2_sb_field_get(c->disk_sb.sb, members_v2);
 	unsigned u64s = DIV_ROUND_UP(sizeof(struct bch_sb_field_members_v2) +
-				     sizeof(struct bch_member), sizeof(u64));
+				     le16_to_cpu(mi->member_bytes), sizeof(u64));
 	bch2_sb_field_resize(&c->disk_sb, members_v2, u64s);
 	c->disk_sb.sb->nr_devices = 1;
 	SET_BCH_SB_MULTI_DEVICE(c->disk_sb.sb, false);
@@ -461,7 +463,7 @@ static void image_create(struct bch_opt_strs	fs_opt_strs,
 
 	struct bch_opts opts = bch2_opts_empty();
 	opt_set(opts, copygc_enabled,		false);
-	opt_set(opts, rebalance_enabled,	false);
+	opt_set(opts, reconcile_enabled,	false);
 	opt_set(opts, nostart,			true);
 
 	struct bch_fs *c = bch2_fs_open(&device_paths, &opts);
@@ -484,10 +486,15 @@ static void image_create(struct bch_opt_strs	fs_opt_strs,
 	struct copy_fs_state s = { .verbosity = verbosity };
 	ret =   copy_fs(c, &s, src_fd, src_path) ?:
 		finish_image(c, keep_alloc, verbosity);
+	bch_err_msg(c, ret, "syncing data");
 	if (ret)
 		goto err;
 
-	bch2_fs_stop(c);
+	ret = bch2_fs_exit(c);
+	if (ret) {
+		fprintf(stderr, "error shutting down new filesystem: %s\n", bch2_err_str(ret));
+		goto err;
+	}
 	darray_exit(&device_paths);
 	xclose(src_fd);
 	return;
@@ -654,7 +661,7 @@ static int image_update(const char *src_path, const char *dst_image,
 
 	struct bch_opts opts = bch2_opts_empty();
 	opt_set(opts, copygc_enabled,		false);
-	opt_set(opts, rebalance_enabled,	false);
+	opt_set(opts, reconcile_enabled,	false);
 	opt_set(opts, nostart,			true);
 
 	struct bch_fs *c = bch2_fs_open(&device_paths, &opts);
@@ -726,18 +733,14 @@ static int image_update(const char *src_path, const char *dst_image,
 
 	ret =   copy_fs(c, &s, src_fd, src_path) ?:
 		finish_image(c, keep_alloc, verbosity);
-	if (ret)
-		goto err;
+	bch_err_msg(c, ret, "syncing data");
 
-	bch2_fs_stop(c);
-	unlink(dev_opts.path);
-	xclose(src_fd);
-	return 0;
 err_stop:
-	bch2_fs_stop(c);
+	ret = bch2_fs_exit(c) ?: ret;
 err:
 	unlink(dev_opts.path);
-	exit(EXIT_FAILURE);
+	xclose(src_fd);
+	return ret ? EXIT_FAILURE : EXIT_SUCCESS;
 }
 
 static void image_update_usage(void)
