@@ -16,6 +16,8 @@
 
 #include "journal/journal.h"
 
+#include "sb/counters.h"
+
 #include <linux/prefetch.h>
 #include <linux/sched/mm.h>
 #include <linux/swap.h>
@@ -34,7 +36,7 @@ void bch2_recalc_btree_reserve(struct bch_fs *c)
 {
 	unsigned reserve = 16;
 
-	if (!c->btree_roots_known[0].b)
+	if (!c->btree.cache.roots_known[0].b)
 		reserve += 8;
 
 	for (unsigned i = 0; i < btree_id_nr_alive(c); i++) {
@@ -44,12 +46,13 @@ void bch2_recalc_btree_reserve(struct bch_fs *c)
 			reserve += min_t(unsigned, 1, r->b->c.level) * 8;
 	}
 
-	c->btree_cache.nr_reserve = reserve;
+	c->btree.cache.nr_reserve = reserve;
 }
 
 static inline size_t btree_cache_can_free(struct btree_cache_list *list)
 {
-	struct btree_cache *bc = container_of(list, struct btree_cache, live[list->idx]);
+	struct bch_fs_btree_cache *bc =
+		container_of(list, struct bch_fs_btree_cache, live[list->idx]);
 
 	size_t can_free = list->nr;
 	if (!list->idx)
@@ -57,7 +60,7 @@ static inline size_t btree_cache_can_free(struct btree_cache_list *list)
 	return can_free;
 }
 
-static void btree_node_to_freedlist(struct btree_cache *bc, struct btree *b)
+static void btree_node_to_freedlist(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	BUG_ON(!list_empty(&b->list));
 
@@ -67,7 +70,7 @@ static void btree_node_to_freedlist(struct btree_cache *bc, struct btree *b)
 		list_add(&b->list, &bc->freed_nonpcpu);
 }
 
-static void __bch2_btree_node_to_freelist(struct btree_cache *bc, struct btree *b)
+static void __bch2_btree_node_to_freelist(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	BUG_ON(!list_empty(&b->list));
 	BUG_ON(!b->data);
@@ -78,7 +81,7 @@ static void __bch2_btree_node_to_freelist(struct btree_cache *bc, struct btree *
 
 void bch2_btree_node_to_freelist(struct bch_fs *c, struct btree *b)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 
 	scoped_guard(mutex, &bc->lock)
 		__bch2_btree_node_to_freelist(bc, b);
@@ -116,7 +119,7 @@ void __btree_node_data_free(struct btree *b)
 	b->aux_data = NULL;
 }
 
-static void btree_node_data_free(struct btree_cache *bc, struct btree *b)
+static void btree_node_data_free(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	BUG_ON(list_empty(&b->list));
 	list_del_init(&b->list);
@@ -201,7 +204,7 @@ struct btree *__bch2_btree_node_mem_alloc(struct bch_fs *c)
 	return b;
 }
 
-static inline bool __btree_node_pinned(struct btree_cache *bc, struct btree *b)
+static inline bool __btree_node_pinned(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	struct bbpos pos = BBPOS(b->c.btree_id, b->key.k.p);
 
@@ -214,7 +217,7 @@ static inline bool __btree_node_pinned(struct btree_cache *bc, struct btree *b)
 
 void bch2_node_pin(struct bch_fs *c, struct btree *b)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 
 	guard(mutex)(&bc->lock);
 	if (!btree_node_is_root(c, b) && !btree_node_pinned(b)) {
@@ -227,12 +230,12 @@ void bch2_node_pin(struct bch_fs *c, struct btree *b)
 
 void bch2_btree_cache_unpin(struct bch_fs *c)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b, *n;
 
 	guard(mutex)(&bc->lock);
-	c->btree_cache.pinned_nodes_mask[0] = 0;
-	c->btree_cache.pinned_nodes_mask[1] = 0;
+	bc->pinned_nodes_mask[0] = 0;
+	bc->pinned_nodes_mask[1] = 0;
 
 	list_for_each_entry_safe(b, n, &bc->live[1].list, list) {
 		clear_btree_node_pinned(b);
@@ -244,13 +247,15 @@ void bch2_btree_cache_unpin(struct bch_fs *c)
 
 /* Btree in memory cache - hash table */
 
-void __bch2_btree_node_hash_remove(struct btree_cache *bc, struct btree *b)
+void __bch2_btree_node_hash_remove(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	lockdep_assert_held(&bc->lock);
 
 	int ret = rhashtable_remove_fast(&bc->table, &b->hash, bch_btree_cache_params);
 	BUG_ON(ret);
 
+	clear_btree_node_just_written(b);
+
 	/* Cause future lookups for this node to fail: */
 	b->hash_val = 0;
 
@@ -260,13 +265,13 @@ void __bch2_btree_node_hash_remove(struct btree_cache *bc, struct btree *b)
 	list_del_init(&b->list);
 }
 
-void bch2_btree_node_hash_remove(struct btree_cache *bc, struct btree *b)
+void bch2_btree_node_hash_remove(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	__bch2_btree_node_hash_remove(bc, b);
 	__bch2_btree_node_to_freelist(bc, b);
 }
 
-int __bch2_btree_node_hash_insert(struct btree_cache *bc, struct btree *b)
+int __bch2_btree_node_hash_insert(struct bch_fs_btree_cache *bc, struct btree *b)
 {
 	BUG_ON(!list_empty(&b->list));
 	BUG_ON(b->hash_val);
@@ -285,7 +290,7 @@ int __bch2_btree_node_hash_insert(struct btree_cache *bc, struct btree *b)
 	return 0;
 }
 
-int bch2_btree_node_hash_insert(struct btree_cache *bc, struct btree *b,
+int bch2_btree_node_hash_insert(struct bch_fs_btree_cache *bc, struct btree *b,
 				unsigned level, enum btree_id id)
 {
 	b->c.level	= level;
@@ -299,7 +304,7 @@ void bch2_btree_node_update_key_early(struct btree_trans *trans,
 				      enum btree_id btree, unsigned level,
 				      struct bkey_s_c old, struct bkey_i *new)
 {
-	struct bch_fs *c = trans->c;
+	struct bch_fs_btree_cache *bc = &trans->c->btree.cache;
 	struct btree *b;
 	struct bkey_buf tmp __cleanup(bch2_bkey_buf_exit);
 	int ret;
@@ -309,12 +314,12 @@ void bch2_btree_node_update_key_early(struct btree_trans *trans,
 
 	b = bch2_btree_node_get_noiter(trans, tmp.k, btree, level, true);
 	if (!IS_ERR_OR_NULL(b)) {
-		guard(mutex)(&c->btree_cache.lock);
+		guard(mutex)(&bc->lock);
 
-		__bch2_btree_node_hash_remove(&c->btree_cache, b);
+		__bch2_btree_node_hash_remove(bc, b);
 
 		bkey_copy(&b->key, new);
-		ret = __bch2_btree_node_hash_insert(&c->btree_cache, b);
+		ret = __bch2_btree_node_hash_insert(bc, b);
 		BUG_ON(ret);
 
 		six_unlock_read(&b->c.lock);
@@ -322,8 +327,8 @@ void bch2_btree_node_update_key_early(struct btree_trans *trans,
 }
 
 __flatten
-static inline struct btree *btree_cache_find(struct btree_cache *bc,
-				     const struct bkey_i *k)
+static inline struct btree *btree_cache_find(struct bch_fs_btree_cache *bc,
+					     const struct bkey_i *k)
 {
 	u64 v = btree_ptr_hash_val(k);
 
@@ -333,7 +338,7 @@ static inline struct btree *btree_cache_find(struct btree_cache *bc,
 static int __btree_node_reclaim_checks(struct bch_fs *c, struct btree *b,
 				       bool flush, bool locked)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 
 	lockdep_assert_held(&bc->lock);
 
@@ -399,7 +404,7 @@ static int __btree_node_reclaim_checks(struct bch_fs *c, struct btree *b,
  */
 static int __btree_node_reclaim(struct bch_fs *c, struct btree *b, bool flush)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	int ret = 0;
 
 	lockdep_assert_held(&bc->lock);
@@ -451,14 +456,15 @@ static unsigned long bch2_btree_cache_scan(struct shrinker *shrink,
 					   struct shrink_control *sc)
 {
 	struct btree_cache_list *list = shrink->private_data;
-	struct btree_cache *bc = container_of(list, struct btree_cache, live[list->idx]);
-	struct bch_fs *c = container_of(bc, struct bch_fs, btree_cache);
+	struct bch_fs_btree_cache *bc =
+		container_of(list, struct bch_fs_btree_cache, live[list->idx]);
+	struct bch_fs *c = container_of(bc, struct bch_fs, btree.cache);
 	struct btree *b, *t;
 	unsigned long nr = sc->nr_to_scan;
 	unsigned long can_free = 0;
 	unsigned long freed = 0;
 	unsigned long touched = 0;
-	unsigned i, flags;
+	unsigned i;
 	unsigned long ret = SHRINK_STOP;
 	bool trigger_writes = atomic_long_read(&bc->nr_dirty) + nr >= list->nr * 3 / 4;
 
@@ -466,7 +472,7 @@ static unsigned long bch2_btree_cache_scan(struct shrinker *shrink,
 		return SHRINK_STOP;
 
 	mutex_lock(&bc->lock);
-	flags = memalloc_nofs_save();
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 
 	/*
 	 * It's _really_ critical that we don't free too many btree nodes - we
@@ -549,8 +555,10 @@ out:
 	mutex_unlock(&bc->lock);
 out_nounlock:
 	ret = freed;
-	memalloc_nofs_restore(flags);
-	trace_and_count(c, btree_cache_scan, sc->nr_to_scan, can_free, ret);
+
+	event_inc_trace(c, btree_cache_scan, buf,
+		prt_printf(&buf, "scanned %li nodes, can free %li, ret %li",
+			   sc->nr_to_scan, can_free, ret));
 	return ret;
 }
 
@@ -567,16 +575,15 @@ static unsigned long bch2_btree_cache_count(struct shrinker *shrink,
 
 void bch2_fs_btree_cache_exit(struct bch_fs *c)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b, *t;
-	unsigned long flags;
 
 	shrinker_free(bc->live[1].shrink);
 	shrinker_free(bc->live[0].shrink);
 
 	/* vfree() can allocate memory: */
-	flags = memalloc_nofs_save();
-	mutex_lock(&bc->lock);
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
+	guard(mutex)(&bc->lock);
 
 	if (c->verify_data)
 		list_move(&c->verify_data->list, &bc->live[0].list);
@@ -603,8 +610,7 @@ void bch2_fs_btree_cache_exit(struct bch_fs *c)
 		cond_resched();
 	}
 
-	BUG_ON(!bch2_journal_error(&c->journal) &&
-	       atomic_long_read(&c->btree_cache.nr_dirty));
+	BUG_ON(!bch2_journal_error(&c->journal) && atomic_long_read(&bc->nr_dirty));
 
 	list_splice(&bc->freed_pcpu, &bc->freed_nonpcpu);
 
@@ -614,9 +620,6 @@ void bch2_fs_btree_cache_exit(struct bch_fs *c)
 		kfree(b);
 	}
 
-	mutex_unlock(&bc->lock);
-	memalloc_nofs_restore(flags);
-
 	for (unsigned i = 0; i < ARRAY_SIZE(bc->nr_by_btree); i++)
 		BUG_ON(bc->nr_by_btree[i]);
 	BUG_ON(bc->live[0].nr);
@@ -625,11 +628,13 @@ void bch2_fs_btree_cache_exit(struct bch_fs *c)
 
 	if (bc->table_init_done)
 		rhashtable_destroy(&bc->table);
+
+	darray_exit(&bc->roots_extra);
 }
 
 int bch2_fs_btree_cache_init(struct bch_fs *c)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct shrinker *shrink;
 
 	if (rhashtable_init(&bc->table, &bch_btree_cache_params))
@@ -673,8 +678,9 @@ int bch2_fs_btree_cache_init(struct bch_fs *c)
 	return 0;
 }
 
-void bch2_fs_btree_cache_init_early(struct btree_cache *bc)
+void bch2_fs_btree_cache_init_early(struct bch_fs_btree_cache *bc)
 {
+	mutex_init(&bc->root_lock);
 	mutex_init(&bc->lock);
 	for (unsigned i = 0; i < ARRAY_SIZE(bc->live); i++) {
 		bc->live[i].idx = i;
@@ -693,11 +699,11 @@ void bch2_fs_btree_cache_init_early(struct btree_cache *bc)
  */
 void bch2_btree_cache_cannibalize_unlock(struct btree_trans *trans)
 {
-	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &trans->c->btree.cache;
 
 	if (bc->alloc_lock == current) {
-		trace_and_count(c, btree_cache_cannibalize_unlock, trans);
+		event_inc_trace(trans->c, btree_cache_cannibalize_unlock, buf,
+			prt_str(&buf, trans->fn));
 		bc->alloc_lock = NULL;
 		closure_wake_up(&bc->alloc_wait);
 	}
@@ -705,7 +711,7 @@ void bch2_btree_cache_cannibalize_unlock(struct btree_trans *trans)
 
 static int __btree_cache_cannibalize_lock(struct bch_fs *c, struct closure *cl)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct task_struct *old;
 
 	old = NULL;
@@ -733,17 +739,15 @@ int bch2_btree_cache_cannibalize_lock(struct btree_trans *trans, struct closure
 	struct bch_fs *c = trans->c;
 	int ret = __btree_cache_cannibalize_lock(c, cl);
 	if (!ret)
-		trace_and_count(c, btree_cache_cannibalize_lock, trans);
-	else if (cl)
-		trace_and_count(c, btree_cache_cannibalize_lock_fail, trans);
+		event_inc_trace(c, btree_cache_cannibalize_lock, buf, prt_str(&buf, trans->fn));
 	else
-		trace_and_count(c, btree_cache_cannibalize_lock_fail, trans);
+		event_inc_trace(c, btree_cache_cannibalize_lock_fail, buf, prt_str(&buf, trans->fn));
 	return ret;
 }
 
 static struct btree *btree_node_cannibalize(struct bch_fs *c)
 {
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b;
 
 	for (unsigned i = 0; i < ARRAY_SIZE(bc->live); i++)
@@ -769,7 +773,7 @@ static struct btree *btree_node_cannibalize(struct bch_fs *c)
 struct btree *bch2_btree_node_mem_alloc(struct btree_trans *trans, bool pcpu_read_locks)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct list_head *freed = pcpu_read_locks
 		? &bc->freed_pcpu
 		: &bc->freed_nonpcpu;
@@ -862,7 +866,6 @@ err:
 	/* Try to cannibalize another cached btree node: */
 	if (bc->alloc_lock == current) {
 		b2 = btree_node_cannibalize(c);
-		clear_btree_node_just_written(b2);
 		__bch2_btree_node_hash_remove(bc, b2);
 
 		if (b) {
@@ -878,7 +881,7 @@ err:
 		BUG_ON(!list_empty(&b->list));
 		mutex_unlock(&bc->lock);
 
-		trace_and_count(c, btree_cache_cannibalize, trans);
+		event_inc_trace(c, btree_cache_cannibalize, buf, prt_str(&buf, trans->fn));
 		goto out;
 	}
 
@@ -896,9 +899,11 @@ static noinline struct btree *bch2_btree_node_fill(struct btree_trans *trans,
 				bool sync)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b;
 
+	EBUG_ON(path && level + 1 != path->level);
+
 	if (unlikely(level >= BTREE_MAX_DEPTH)) {
 		int ret = bch2_fs_topology_error(c, "attempting to get btree node at level %u, >= max depth %u",
 						 level, BTREE_MAX_DEPTH);
@@ -925,9 +930,10 @@ static noinline struct btree *bch2_btree_node_fill(struct btree_trans *trans,
 	 * Parent node must be locked, else we could read in a btree node that's
 	 * been freed:
 	 */
-	if (path && !bch2_btree_node_relock(trans, path, level + 1)) {
-		trace_and_count(c, trans_restart_relock_parent_for_fill, trans, _THIS_IP_, path);
-		return ERR_PTR(btree_trans_restart(trans, BCH_ERR_transaction_restart_fill_relock));
+	if (path) {
+		int ret = bch2_btree_path_relock(trans, path, _THIS_IP_);
+		if (ret)
+			return ERR_PTR(ret);
 	}
 
 	b = bch2_btree_node_mem_alloc(trans, level != 0);
@@ -937,7 +943,11 @@ static noinline struct btree *bch2_btree_node_fill(struct btree_trans *trans,
 			return b;
 
 		trans->memory_allocation_failure = true;
-		trace_and_count(c, trans_restart_memory_allocation_failure, trans, _THIS_IP_, path);
+
+		event_inc_trace(c, trans_restart_memory_allocation_failure, buf, ({
+			prt_printf(&buf, "%s\n", trans->fn);
+			bch2_btree_path_to_text(&buf, trans, path - trans->paths, path);
+		}));
 		return ERR_PTR(btree_trans_restart(trans, BCH_ERR_transaction_restart_fill_mem_alloc_fail));
 	}
 
@@ -972,7 +982,8 @@ static noinline struct btree *bch2_btree_node_fill(struct btree_trans *trans,
 
 		bch2_btree_node_read(trans, b, sync);
 
-		int ret = bch2_trans_relock(trans);
+		int ret = bch2_trans_relock(trans) ?:
+			  bch2_btree_path_relock(trans, path, _THIS_IP_);
 		if (ret)
 			return ERR_PTR(ret);
 
@@ -1030,9 +1041,8 @@ static struct btree *__bch2_btree_node_get(struct btree_trans *trans, struct btr
 					   unsigned long trace_ip)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b;
-	bool need_relock = false;
 	int ret;
 
 	EBUG_ON(level >= BTREE_MAX_DEPTH);
@@ -1046,7 +1056,6 @@ retry:
 		 */
 		b = bch2_btree_node_fill(trans, path, k, path->btree_id,
 					 level, lock_type, true);
-		need_relock = true;
 
 		/* We raced and found the btree node in the cache */
 		if (!b)
@@ -1071,7 +1080,10 @@ retry:
 			if (bch2_btree_node_relock(trans, path, level + 1))
 				goto retry;
 
-			trace_and_count(c, trans_restart_btree_node_reused, trans, trace_ip, path);
+			event_inc_trace(c, trans_restart_btree_node_reused, buf, ({
+				prt_printf(&buf, "%s\n", trans->fn);
+				bch2_btree_path_to_text(&buf, trans, path - trans->paths, path);
+			}));
 			return ERR_PTR(btree_trans_restart(trans, BCH_ERR_transaction_restart_lock_node_reused));
 		}
 
@@ -1085,11 +1097,11 @@ retry:
 
 		six_unlock_type(&b->c.lock, lock_type);
 		bch2_trans_unlock(trans);
-		need_relock = true;
 
 		bch2_btree_node_wait_on_read(b);
 
-		ret = bch2_trans_relock(trans);
+		ret =   bch2_trans_relock(trans) ?:
+			bch2_btree_path_relock(trans, path, _THIS_IP_);
 		if (ret)
 			return ERR_PTR(ret);
 
@@ -1101,15 +1113,6 @@ retry:
 			goto retry;
 	}
 
-	if (unlikely(need_relock)) {
-		ret = bch2_trans_relock(trans) ?:
-			bch2_btree_path_relock_intent(trans, path);
-		if (ret) {
-			six_unlock_type(&b->c.lock, lock_type);
-			return ERR_PTR(ret);
-		}
-	}
-
 	prefetch(b->aux_data);
 
 	for_each_bset(b, t) {
@@ -1158,6 +1161,7 @@ struct btree *bch2_btree_node_get(struct btree_trans *trans, struct btree_path *
 	int ret;
 
 	EBUG_ON(level >= BTREE_MAX_DEPTH);
+	EBUG_ON(level + 1 != path->level);
 
 	b = btree_node_mem_ptr(k);
 
@@ -1187,7 +1191,10 @@ struct btree *bch2_btree_node_get(struct btree_trans *trans, struct btree_path *
 		if (bch2_btree_node_relock(trans, path, level + 1))
 			return __bch2_btree_node_get(trans, path, k, level, lock_type, trace_ip);
 
-		trace_and_count(c, trans_restart_btree_node_reused, trans, trace_ip, path);
+		event_inc_trace(c, trans_restart_btree_node_reused, buf, ({
+			prt_printf(&buf, "%s\n", trans->fn);
+			bch2_btree_path_to_text(&buf, trans, path - trans->paths, path);
+		}));
 		return ERR_PTR(btree_trans_restart(trans, BCH_ERR_transaction_restart_lock_node_reused));
 	}
 
@@ -1229,7 +1236,7 @@ struct btree *bch2_btree_node_get_noiter(struct btree_trans *trans,
 					 bool nofill)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b;
 	int ret;
 
@@ -1312,7 +1319,7 @@ int bch2_btree_node_prefetch(struct btree_trans *trans,
 			     enum btree_id btree_id, unsigned level)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 
 	BUG_ON(path && !btree_node_locked(path, level + 1));
 	BUG_ON(level >= BTREE_MAX_DEPTH);
@@ -1331,7 +1338,7 @@ int bch2_btree_node_prefetch(struct btree_trans *trans,
 void bch2_btree_node_evict(struct btree_trans *trans, const struct bkey_i *k)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_cache *bc = &c->btree_cache;
+	struct bch_fs_btree_cache *bc = &c->btree.cache;
 	struct btree *b;
 
 	b = btree_cache_find(bc, k);
@@ -1446,7 +1453,7 @@ void bch2_btree_node_to_text(struct printbuf *out, struct bch_fs *c, const struc
 	       b->nr.live_u64s * 100 / btree_max_u64s(c),
 	       b->sib_u64s[0],
 	       b->sib_u64s[1],
-	       c->btree_foreground_merge_threshold,
+	       c->btree.foreground_merge_threshold,
 	       b->nr.packed_keys,
 	       b->nr.unpacked_keys,
 	       stats.floats,
@@ -1468,9 +1475,9 @@ static const char * const bch2_btree_cache_not_freed_reasons_strs[] = {
 	NULL
 };
 
-void bch2_btree_cache_to_text(struct printbuf *out, const struct btree_cache *bc)
+void bch2_btree_cache_to_text(struct printbuf *out, const struct bch_fs_btree_cache *bc)
 {
-	struct bch_fs *c = container_of(bc, struct bch_fs, btree_cache);
+	struct bch_fs *c = container_of(bc, struct bch_fs, btree.cache);
 
 	if (!out->nr_tabstops)
 		printbuf_tabstop_push(out, 32);
